# -*- coding: utf-8 -*-

"""Provider for SagePub

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class SagePubInformer(Informer):

    meta_author_key = "dc.Creator"
    meta_title_key = "dc.Title"
    meta_date_key = "dc.Date"

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)

    def _format_year(self, soup_date):
        return soup_date.split("-")[0]


class SagePub(Provider):

    re_abs = "https?:\/\/journals\.sagepub\.com\/doi\/full\/\d{2}\.\d{4}\/\d+"
    re_pdf = "https?:\/\/journals\.sagepub\.com\/doi\/pdf\/\d{2}\.\d{4}\/\d+"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = SagePubInformer()

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url.replace("full", "pdf")
        elif re.match(self.re_pdf, url):
            pdf_url = url
            abs_url = url.replace("pdf", "full")
        else:
            raise URLResolutionError("SagePub", url)
        return abs_url, pdf_url

    def validate(src):
        return re.match(SagePub.re_abs, src) or re.match(SagePub.re_pdf, src)



# -*- coding: utf-8 -*-

"""Provider for CVF

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer

from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()


class CVFInformer(Informer):

    meta_date_key = "citation_publication_date"


class CVF(Provider):

    re_abs = "^https?://openaccess.thecvf.com/content_([\w\d]+)/html/([\w\d\_\-]+).html$"
    re_pdf = "^https?://openaccess.thecvf.com/content_([\w\d]+)/papers/([\w\d\_\-]+).pdf$"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = CVFInformer()

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url[: -len(".html")]
            pdf_url += ".pdf"
            pdf_url = pdf_url.replace("html", "papers")
        elif re.match(self.re_pdf, url):
            pdf_url = url
            abs_url = url.replace("papers", "html").replace(".pdf", ".html")
        else:
            raise URLResolutionError("CVF", url)
        return abs_url, pdf_url

    def validate(src):
        m = re.match(CVF.re_abs, src) or re.match(CVF.re_pdf, src)
        return not m is None



# -*- coding: utf-8 -*-

"""Provider for SemanticScholar

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re
import bs4

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..utils import get_page_with_retry


class SemanticScholarInformer(Informer):

    meta_date_key = "citation_publication_date"

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class SemanticScholar(Provider):

    re_abs = (
        "https?:\/\/www.semanticscholar.org/paper/[A-Za-z0-9%\-]+/[0-9a-f]{40}"
    )
    re_pdf = "https?:\/\/pdfs.semanticscholar.org/[0-9a-f]{4}/[0-9a-f]{36}.pdf"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = SemanticScholarInformer()

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract urls from a SemanticScholar url """
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = self._get_pdf_url(abs_url)
        elif re.match(self.re_pdf, url):
            pdf_url = url
            remainder = pdf_url.split("/")[-1][: -len(".pdf")]
            first_four = pdf_url.split("/")[-2]
            paper_id = first_four + remainder
            abs_url = f"https://www.semanticscholar.org/paper/{paper_id}"
        else:
            raise URLResolutionError("SemanticScholar", url)
        return abs_url, pdf_url

    def _get_pdf_url(self, url):
        page = get_page_with_retry(url)
        soup = bs4.BeautifulSoup(page, "html.parser")
        meta = soup.find_all("meta", {"name": "citation_pdf_url"})
        if not meta:
            raise URLResolutionError("SemanticScholar", url)
        return meta[0]["content"]

    def validate(src):
        return re.match(SemanticScholar.re_abs, src) or re.match(
            SemanticScholar.re_pdf, src
        )



# -*- coding: utf-8 -*-

"""Provider for Springer

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re
import urllib
import requests

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..utils import HEADERS


class SpringerInformer(Informer):

    meta_date_key = None

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)

    def get_year(self, soup):
        for key in ["citation_online_date", "citation_publication_date"]:
            meta = soup.find_all("meta", {"name": key})
            if not meta:
                continue
            return self._format_year(meta[0]["content"])
        return ""


class Springer(Provider):

    re_abs_1 = "https?:\/\/link.springer.com\/article\/10\.\d{4}\/[a-z0-9\-]+"
    re_abs_2 = "https?:\/\/link.springer.com\/chapter\/10\.\d{4}\/[a-z0-9\-]+"
    re_pdf = "https?:\/\/link\.springer\.com\/content\/pdf\/10\.\d{4}(%2F|\/)[a-z0-9\-\_]+\.pdf"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = SpringerInformer()

    def _get_abs_url(self, pdf_url):
        article_url = pdf_url.replace("content/pdf", "article")[: -len(".pdf")]
        req = requests.head(
            article_url, headers=HEADERS, cookies=self.cookiejar
        )
        if req.status_code == 200:
            return article_url

        chapter_url = pdf_url.replace("content/pdf", "chapter")[: -len(".pdf")]
        req = requests.head(
            chapter_url, headers=HEADERS, cookies=self.cookiejar
        )
        if req.status_code == 200:
            return chapter_url

        raise URLResolutionError("Springer", pdf_url)

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract urls from a Springer url """
        if re.match(self.re_abs_1, url):
            abs_url = url
            pdf_url = url.replace("article", "content/pdf")
        elif re.match(self.re_abs_2, url):
            abs_url = url
            pdf_url = url.replace("chapter", "content/pdf")
        elif re.match(self.re_pdf, url):
            abs_url = self._get_abs_url(url)
            pdf_url = urllib.parse.unquote(url)
        else:
            raise URLResolutionError("Springer", url)
        return abs_url, pdf_url

    def validate(src):
        return (
            re.match(Springer.re_abs_1, src)
            or re.match(Springer.re_abs_2, src)
            or re.match(Springer.re_pdf, src)
        )



# -*- coding: utf-8 -*-

"""Provider for OpenReview

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import json
import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()


class OpenReviewInformer(Informer):

    meta_date_key = "citation_publication_date"

    def get_authors(self, soup):
        # Get the authors for OpenReview by parsing the JSON payload
        #
        # This may not be super robust long term, but works for now.
        warning = (
            "Couldn't determine author information, maybe provide "
            "the desired filename using '--filename'?"
        )

        script = soup.find("script", {"id": "__NEXT_DATA__"})
        if not script:
            logger.warning(warning)
            return ""

        try:
            paper_data = json.loads(script.contents[0])
        except json.JSONDecodeError:
            logger.warning(warning)
            return ""

        try:
            content = paper_data["props"]["pageProps"]["forumNote"]["content"]
            authors = content["authors"]
        except KeyError:
            logger.warning(warning)
            return ""
        return self._format_authors(authors)

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class OpenReview(Provider):

    re_abs = "https?://openreview.net/forum\?id=[A-Za-z0-9]+"
    re_pdf = "https?://openreview.net/pdf\?id=[A-Za-z0-9]+"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = OpenReviewInformer()

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract url from a OpenReview url """
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url.replace("forum", "pdf")
        elif re.match(self.re_pdf, url):
            abs_url = url.replace("pdf", "forum")
            pdf_url = url
        else:
            raise URLResolutionError("OpenReview", url)
        return abs_url, pdf_url

    def validate(src):
        """ Check if the url is a valid OpenReview url. """
        return re.match(OpenReview.re_abs, src) or re.match(
            OpenReview.re_pdf, src
        )



# -*- coding: utf-8 -*-

"""Provider for JMLR

Journal of Machine Learning Research

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class JMLRInformer(Informer):

    meta_date_key = "citation_publication_date"

    def _format_authors(self, soup_authors):
        have_comma = any(("," in auth for auth in soup_authors))
        if have_comma:
            return super()._format_authors(soup_authors, sep=",", idx=0)
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class JMLR(Provider):

    re_abs_1 = "https?://(www\.)?jmlr\.org/papers/v(?P<vol>\d+)/(?P<pid>\d{2}\-\d{3}).html$"
    re_pdf_1 = "https?://(www\.)?jmlr\.org/papers/volume(?P<vol>\d+)/(?P<pid>\d{2}\-\d{3})/(?P=pid).pdf$"

    re_abs_2 = "https?://(www\.)?jmlr\.org/papers/v(?P<vol>\d+)/(?P<pid>\w+\d{2}\w).html$"
    re_pdf_2 = "https?://(www\.)?jmlr\.org/papers/volume(?P<vol>\d+)/(?P<pid>\w+\d{2}\w)/(?P=pid).pdf$"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = JMLRInformer()

    def get_abs_pdf_urls(self, url):
        abs_url = pdf_url = None
        abs_fmt = "http://jmlr.org/papers/v{vol}/{pid}.html"
        pdf_fmt = "http://jmlr.org/papers/volume{vol}/{pid}/{pid}.pdf"
        formats = [
            (self.re_abs_1, self.re_pdf_1),
            (self.re_abs_2, self.re_pdf_2),
        ]

        for re_abs, re_pdf in formats:
            ma = re.match(re_abs, url)
            mp = re.match(re_pdf, url)
            if ma:
                abs_url = url
                pdf_url = pdf_fmt.format(
                    vol=ma.group("vol"), pid=ma.group("pid")
                )
            elif mp:
                abs_url = abs_fmt.format(
                    vol=mp.group("vol"), pid=mp.group("pid")
                )
                pdf_url = url
        if abs_url is None or pdf_url is None:
            raise URLResolutionError("JMLR", url)
        return abs_url, pdf_url

    def validate(src):
        return (
            re.match(JMLR.re_abs_1, src)
            or re.match(JMLR.re_abs_2, src)
            or re.match(JMLR.re_pdf_1, src)
            or re.match(JMLR.re_pdf_2, src)
        )



# -*- coding: utf-8 -*-

"""Provider for NeurIPS

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class NeurIPSInformer(Informer):

    meta_date_key = "citation_publication_date"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_site = False

    def _format_authors(self, soup_authors):
        if self.new_site:
            return super()._format_authors(soup_authors, sep=",", idx=0)
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class NeurIPS(Provider):

    re_abs = "^https?://papers.n(eur)?ips.cc/paper/[\d\w\-]+$"
    re_pdf = "^https?://papers.n(eur)?ips.cc/paper/[\d\w\-]+.pdf$"

    re_abs_2 = "https://papers.n(eur)?ips.cc/paper/\d{4}/hash/[0-9a-f]{32}-Abstract.html"
    re_pdf_2 = (
        "https://papers.n(eur)?ips.cc/paper/\d{4}/file/[0-9a-f]{32}-Paper.pdf"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = NeurIPSInformer()

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract url from a OpenReview url """
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url + ".pdf"
        elif re.match(self.re_pdf, url):
            abs_url = url.replace(".pdf", "")
            pdf_url = url
        elif re.match(self.re_abs_2, url):
            self.informer.new_site = True
            abs_url = url
            pdf_url = (
                url.replace("hash", "file")
                .replace("Abstract", "Paper")
                .replace(".html", ".pdf")
            )
        elif re.match(self.re_pdf_2, url):
            self.informer.new_site = True
            pdf_url = url
            abs_url = (
                url.replace("file", "hash")
                .replace("Paper", "Abstract")
                .replace(".pdf", ".html")
            )
        else:
            raise URLResolutionError("NeurIPS", url)
        return abs_url, pdf_url

    def validate(src):
        return (
            re.fullmatch(NeurIPS.re_abs, src)
            or re.fullmatch(NeurIPS.re_pdf, src)
            or re.fullmatch(NeurIPS.re_abs_2, src)
            or re.fullmatch(NeurIPS.re_pdf_2, src)
        )



# -*- coding: utf-8 -*-

"""Code for cropping a PDF file

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import PyPDF2
import io
import os
import pdfplumber
import subprocess

from PyPDF2.generic import RectangleObject

from .log import Logger

RM_WIDTH = 1404
RM_HEIGHT = 1872

logger = Logger()


def find_offset_byte_line(line):
    """Find index of first nonzero bit in a line of bytes

    The given line is a string of bytes, each representing 8 pixels. This code
    finds the index of the first bit that is not zero. Used when finding the
    cropbox with pdftoppm.
    """
    off = 0
    for c in line:
        if c == 0:
            off += 8
        else:
            k = 0
            while c > 0:
                k += 1
                c >>= 1
            off += k
            break
    return off


def check_pdftoppm(pth):
    """Check that we can run the provided pdftoppm executable"""
    try:
        subprocess.check_output([pth, "-v"], stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, FileNotFoundError, PermissionError):
        logger.info("pdftoppm not found, using pdfplumber instead (slower)")
        return False
    return True


class Cropper(object):
    def __init__(
        self,
        input_file=None,
        output_file=None,
        pdftoppm_path="pdftoppm",
    ):
        if not input_file is None:
            self.input_file = os.path.abspath(input_file)
            self.reader = PyPDF2.PdfFileReader(self.input_file)
        if not output_file is None:
            self.output_file = os.path.abspath(output_file)

        if pdftoppm_path and not check_pdftoppm(pdftoppm_path):
            pdftoppm_path = None

        self.pdftoppm_path = pdftoppm_path
        self.writer = PyPDF2.PdfFileWriter()

    def crop(self, margins=1):
        return self.process_file(self.crop_page, margins=margins)

    def center(self, padding=15):
        return self.process_file(self.center_page, padding=padding)

    def right(self, padding=15):
        return self.process_file(self.right_page, padding=padding)

    def process_file(self, page_func, *args, **kwargs):
        n = self.reader.getNumPages()
        for page_idx in range(n):
            status = page_func(page_idx, *args, **kwargs)
            if not status == 0:
                return status
            if (page_idx + 1) % 10 == 0:
                logger.info("Processing pages ... (%i/%i)" % (page_idx + 1, n))
        with open(self.output_file, "wb") as fp:
            self.writer.write(fp)
        if n % 10 > 0:
            logger.info("Processing pages ... (%i/%i)" % (n, n))
        return 0

    def crop_page(self, page_idx, margins):
        return self.process_page(page_idx, self.get_bbox, margins=margins)

    def center_page(self, page_idx, padding):
        return self.process_page(
            page_idx, self.get_center_bbox, padding=padding
        )

    def right_page(self, page_idx, padding):
        return self.process_page(
            page_idx, self.get_right_bbox, padding=padding
        )

    def export_page(self, page_idx):
        """Helper function that exports a single page given by index """
        page = self.reader.getPage(page_idx)
        writer = PyPDF2.PdfFileWriter()
        writer.addPage(page)
        tmpfname = "./page.pdf"
        with open(tmpfname, "wb") as fp:
            writer.write(fp)
        return tmpfname

    def process_page(self, page_idx, bbox_func, *args, **kwargs):
        """Process a single page and add it to the writer """
        tmpfname = self.export_page(page_idx)
        bbox = bbox_func(tmpfname, *args, **kwargs)
        thepage = self.reader.getPage(page_idx)
        thepage.cropBox = RectangleObject(bbox)
        self.writer.addPage(thepage)
        os.unlink(tmpfname)
        return 0

    def get_raw_bbox(self, filename, resolution=72):
        """Get the basic bounding box of a pdf file"""
        if self.pdftoppm_path is None:
            box = self.get_raw_bbox_pdfplumber(filename, resolution=resolution)
        else:
            box = self.get_raw_bbox_pdftoppm(filename, resolution=resolution)
        return box

    def get_raw_bbox_pdfplumber(self, filename, resolution=72):
        """Get the basic bounding box with pdfplumber"""
        pdf = pdfplumber.open(filename)
        im = pdf.pages[0].to_image(resolution=resolution)
        pdf.close()

        pixels = list(im.original.getdata())
        W, H = im.original.size

        # M is a list of H lists with each W integers that equal the sum of the
        # pixel values
        M = [[sum(x) for x in pixels[i * W : (i + 1) * W]] for i in range(H)]

        left, top, bottom, right = 0, 0, 0, 0
        while top < H and sum(M[top]) == W * 255 * 3:
            top += 1
        while bottom < H and sum(M[H - 1 - bottom]) == W * 255 * 3:
            bottom += 1

        # Transpose M
        M = list(zip(*M))
        while left < W and sum(M[left]) == H * 255 * 3:
            left += 1
        while right < W and sum(M[W - 1 - right]) == H * 255 * 3:
            right += 1

        return left, right, top, bottom, W, H

    def get_raw_bbox_pdftoppm(self, filename, resolution=72):
        """Get the basic bounding box using pdftoppm """
        cmd = [
            self.pdftoppm_path,
            "-r",
            str(resolution),
            "-singlefile",
            "-mono",
            filename,
        ]

        im = subprocess.check_output(cmd)
        im = io.BytesIO(im)

        id_ = im.readline().rstrip(b"\n")
        if not id_ == b"P4":
            raise ValueError("Not in P4 format")
        wh = im.readline().rstrip(b"\n").split(b" ")
        width, height = int(wh[0]), int(wh[1])
        imdata = im.read()

        pad = width % 8
        padwidth = width + pad
        stepsize = padwidth // 8

        for top in range(height):
            if sum(imdata[top * stepsize : (top + 1) * stepsize]) > 0:
                break

        for bottom in reversed(range(height)):
            if sum(imdata[bottom * stepsize : (bottom + 1) * stepsize]) > 0:
                break

        left = width
        right = 0
        for i in range(top, bottom):
            lline = imdata[i * stepsize : (i + 1) * stepsize]
            rline = reversed(imdata[i * stepsize : (i + 1) * stepsize])
            l = find_offset_byte_line(lline)
            left = min(left, l)
            r = padwidth + pad - find_offset_byte_line(rline)
            right = max(right, r)

        top += 1
        left += 1
        right = width - right + 2
        bottom = height - bottom - 2

        return left, right, top, bottom, width, height

    def get_bbox(self, filename, margins=1, resolution=72):
        """Get the bounding box, with optional margins

        if margins is integer, used for all margins, else
        margins = [left, top, right, bottom]

        We get the bounding box by finding the smallest rectangle that is
        completely surrounded by white pixels.
        """
        if isinstance(margins, int):
            margins = [margins for _ in range(4)]

        left, right, top, bottom, W, H = self.get_raw_bbox(
            filename, resolution=resolution
        )

        left -= margins[0]
        left = max(left, 0)
        top -= margins[1]
        top = max(top, 0)
        right -= margins[2]
        bottom -= margins[3]

        # This is the bounding box in PIL format: (0, 0) top left
        x0, y0, x1, y1 = left, top, W - right, H - bottom

        # The remarkable changes the orientation of a portrait page if the
        # width is greater than the height. To prevent this, we pad the height
        # with extra whitespace. This should only occur if the original
        # orientation of the page would be changed by cropping.
        w, h = x1 - x0, y1 - y0
        if H > W and w > h:
            y1 = y0 + w + 10
            h = y1 - y0

        # Get the bbox in Ghostscript format: (0, 0) bottom left
        a0, b0, a1, b1 = x0, H - y1, x1, H - y0
        return [a0, b0, a1, b1]

    def get_center_bbox(self, filename, padding=15):
        """Compute a bounding box that will center the page file on the
        reMarkable
        """
        bbox = self.get_bbox(filename, margins=0)

        h = bbox[3] - bbox[1]
        w = bbox[2] - bbox[0]

        # we want some minimal padding all around, because it is visually more
        # pleasing.
        h_prime = h + 2 * padding
        w_prime = w + 2 * padding

        # if the document is wider than the remarkable, we add top-padding to
        # center it, otherwise we add left-padding
        x = y = 0
        if h_prime / w_prime < RM_HEIGHT / RM_WIDTH:
            y = ((RM_HEIGHT / RM_WIDTH) * w_prime - h_prime) / 2
        else:
            x = ((RM_WIDTH / RM_HEIGHT) * h_prime - w_prime) / 2

        margins = [padding + x, padding + y, padding, padding]
        return self.get_bbox(filename, margins=margins)

    def get_right_bbox(self, filename, padding=15):
        """Get the bounding box that ensures the menu doesn't hide the text"""

        bbox = self.get_bbox(filename, margins=0)

        h = bbox[3] - bbox[1]
        w = bbox[2] - bbox[0]

        # Note, the menu width is about 12mm and the entire screen is about
        # 156mm. This informs the width of the left padding we'll add.
        menu_width = 12 / 156 * RM_WIDTH

        H = RM_HEIGHT
        W = RM_WIDTH

        # TODO: This math is approximate. The goal is to get the page centered
        # in the remaining space after taking the menu width into account,
        # while also providing equal padding at the top and bottom. This seems
        # to give too much padding on the left for some pages, but I'm not sure
        # why. Pull requests welcome!
        rho_rm = H / (W - menu_width)
        rho_page = (h + 2 * padding) / (w + 2 * padding)
        x = y = 0
        if rho_rm < rho_page:
            x = -w - 2 * padding + (h + 2 * padding) * (W - menu_width) / H
        elif rho_rm > rho_page:
            y = -h - 2 * padding + H * (w + 2 * padding) / (W - menu_width)

        margins = [
            menu_width + x + padding,
            padding + y,
            padding,
            padding,
        ]
        return self.get_bbox(filename, margins=margins)



# -*- coding: utf-8 -*-

"""Utility functions for a2r

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import PyPDF2
import regex
import requests
import string
import subprocess
import time
import unidecode

from .log import Logger
from .exceptions import FileTypeError, RemarkableError, NoPDFToolError

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 "
    "Safari/537.36"
}


logger = Logger()


def clean_string(s):
    """Clean a string by replacing accented characters with equivalents and
    keeping only the allowed characters (ascii letters, digits, underscore,
    space, dash, and period)"""
    normalized = unidecode.unidecode(s)
    allowed = string.ascii_letters + string.digits + "_ .-"
    cleaned = "".join(c if c in allowed else "_" for c in normalized)
    while "__" in cleaned:
        cleaned = cleaned.replace("__", "_")
    cleaned = cleaned.strip("_")
    return cleaned


def assert_file_is_pdf(filename):
    """Assert that a given file is a PDF file.

    This is done by trying to open it using PyPDF2.
    """
    try:
        fp = open(filename, "rb")
        pdf = PyPDF2.PdfFileReader(fp, strict=False)
        fp.close()
        del pdf
        return True
    except PyPDF2.utils.PdfReadError:
        raise FileTypeError(filename, "pdf")


def download_url(url, filename, cookiejar=None):
    """Download the content of an url and save it to a filename """
    logger.info("Downloading file at url: %s" % url)
    content = get_page_with_retry(url, cookiejar=cookiejar)
    with open(filename, "wb") as fid:
        fid.write(content)


def get_page_with_retry(url, tries=5, cookiejar=None, return_text=False):
    count = 0
    jar = {} if cookiejar is None else cookiejar
    while count < tries:
        count += 1
        error = False
        try:
            res = requests.get(url, headers=HEADERS, cookies=jar)
        except requests.exceptions.ConnectionError:
            error = True
        if error or not res.ok:
            logger.warning(
                "(%i/%i) Error getting url %s. Retrying in 5 seconds."
                % (count, tries, url)
            )
            time.sleep(5)
            continue
        logger.info("Downloaded url: %s" % url)
        if return_text:
            return res.text
        return res.content


def get_content_type_with_retry(url, tries=5, cookiejar=None):
    if cookiejar is None:
        jar = requests.cookies.RequestsCookieJar()
    else:
        jar = cookiejar

    msg = "(%i/%i) Error getting content type for %s. Retrying in 5 seconds."

    # In rare cases, a HEAD request fails but a GET request does work. So here
    # we try both
    ops = [requests.head, requests.get]
    kwargs = dict(headers=HEADERS, cookies=jar, allow_redirects=True)
    for op in ops:
        count = 0
        while count < tries:
            count += 1
            error = False
            try:
                res = op(url, **kwargs)
            except requests.exceptions.ConnectionError:
                error = True
            if error or not res.ok:
                logger.warning(msg % (count, tries, url))
                time.sleep(5)
                continue
            return res.headers.get("Content-Type", None)
    return None


def follow_redirects(url):
    """Follow redirects from the URL (at most 100)"""
    it = 0
    jar = requests.cookies.RequestsCookieJar()
    while it < 100:
        req = requests.head(
            url, headers=HEADERS, allow_redirects=False, cookies=jar
        )
        if req.status_code == 200:
            break
        if not "Location" in req.headers:
            break
        url = req.headers["Location"]
        jar.update(req.cookies)
        it += 1
    if it == 100:
        logger.warning("Max redirects reached. There may be a problem.")
    jar = jar or req.cookies
    return url, jar


def upload_to_remarkable(filepath, remarkable_dir="/", rmapi_path="rmapi"):
    logger.info("Starting upload to reMarkable")

    # Create the reMarkable dir if it doesn't exist
    remarkable_dir = remarkable_dir.rstrip("/")
    if remarkable_dir:
        parts = remarkable_dir.split("/")
        rmdir = ""
        while parts:
            rmdir += "/" + parts.pop(0)
            status = subprocess.call(
                [rmapi_path, "mkdir", rmdir],
                stdout=subprocess.DEVNULL,
            )
            if not status == 0:
                raise RemarkableError(
                    "Creating directory %s on reMarkable failed"
                    % remarkable_dir
                )

    # Upload the file
    status = subprocess.call(
        [rmapi_path, "put", filepath, remarkable_dir + "/"],
        stdout=subprocess.DEVNULL,
    )
    if not status == 0:
        raise RemarkableError(
            "Uploading file %s to reMarkable failed" % filepath
        )
    logger.info("Upload successful.")


def is_url(string):
    # pattern adapted from CleverCSV
    pattern = "((https?|ftp):\/\/(?!\-))?(((([\p{L}\p{N}]*[\-\_]?[\p{L}\p{N}]+)+\.)+([a-z]{2,}|local)(\.[a-z]{2,3})?)|localhost|(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(\:\d{1,5})?))(\/[\p{L}\p{N}_\/()~?=&%\-\#\.:+]*)?(\.[a-z]+)?"
    string = string.strip(" ")
    match = regex.fullmatch(pattern, string)
    return match is not None


def check_pdftool(pdftk_path, qpdf_path):
    """Check whether we have pdftk or qpdf available"""
    # set defaults in case either is set to None or something
    pdftk_path = pdftk_path or "false"
    qpdf_path = qpdf_path or "false"

    try:
        status = subprocess.call(
            [pdftk_path], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
    except FileNotFoundError:
        status = 1
    if status == 0:
        return "pdftk"
    try:
        status = subprocess.call(
            [qpdf_path, "--help"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        status = 1
    if status == 0:
        return "qpdf"
    raise NoPDFToolError



# -*- coding: utf-8 -*-

"""Provider for arxiv.org

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import os
import re

from ._info import Informer
from ._base import Provider
from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()

DEARXIV_TEXT_REGEX = b"ar(x|X)iv:(\d{4}\.|[\w\-]+\/)\d+v\d+(\s+\[[\w\-]+\.[\w\-]+\])?\s+\d{1,2}\s\w{3}\s\d{4}"
DEARXIV_URI_REGEX = (
    b"https?://ar(x|X)iv\.org\/abs\/([\w\-]+\/\d+|\d{4}\.\d{4,5})v\d+"
)


class ArxivInformer(Informer):
    pass


class Arxiv(Provider):

    re_abs_1 = "https?://arxiv.org/abs/\d{4}\.\d{4,5}(v\d+)?"
    re_pdf_1 = "https?://arxiv.org/pdf/\d{4}\.\d{4,5}(v\d+)?\.pdf"

    re_abs_2 = "https?://arxiv.org/abs/[\w\-]+/\d{7}(v\d+)?"
    re_pdf_2 = "https?://arxiv.org/pdf/[\w\-]+/\d{7}(v\d+)?.pdf"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = ArxivInformer()

        # register the dearxiv operation
        self.operations.insert(0, ("dearxiv", self.dearxiv))

    def get_abs_pdf_urls(self, url):
        """Get the pdf and abs url from any given arXiv url """
        if "?" in url:
            url = url[: url.index("?")]
        if re.match(self.re_abs_1, url) or re.match(self.re_abs_2, url):
            abs_url = url
            pdf_url = url.replace("abs", "pdf") + ".pdf"
        elif re.match(self.re_pdf_1, url) or re.match(self.re_pdf_2, url):
            abs_url = url[:-4].replace("pdf", "abs")
            pdf_url = url
        else:
            raise URLResolutionError("arXiv", url)
        return abs_url, pdf_url

    def validate(src):
        """Check if the url is to an arXiv page. """
        return (
            re.match(Arxiv.re_abs_1, src)
            or re.match(Arxiv.re_pdf_1, src)
            or re.match(Arxiv.re_abs_2, src)
            or re.match(Arxiv.re_pdf_2, src)
        )

    def dearxiv(self, input_file):
        """Remove the arXiv timestamp from a pdf"""
        logger.info("Removing arXiv timestamp ... ", end="")
        basename = os.path.splitext(input_file)[0]

        recoded_file = basename + "_rewrite.pdf"
        self.rewrite_pdf(input_file, recoded_file)

        uncompress_file = basename + "_uncompress.pdf"
        self.uncompress_pdf(recoded_file, uncompress_file)

        new_data = []
        current_obj = []
        replaced_arXiv = False
        char_count = skip_n = startxref = 0
        xref = {}

        with open(uncompress_file, "rb") as fp:
            for line in fp:
                if skip_n:
                    # Skip a line
                    skip_n -= 1
                    continue

                if line.endswith(b" obj\n") or line.endswith(b" obj \n"):
                    # Start a new object. Add it to the current object and
                    # record its position for the xref table.
                    current_obj.append(line)
                    objid = int(line.split(b" ")[0])
                    xref[objid] = char_count
                elif current_obj and (
                    line.startswith(b"endobj")
                    and not line.startswith(b"endobj xref")
                ):
                    # End the current object. If needed, replace the arXiv
                    # stamp in the block (done only once). Reset current
                    # object.
                    current_obj.append(line)
                    block = b"".join(current_obj)
                    # remove the text
                    block, n_subs1 = re.subn(
                        b"\(" + DEARXIV_TEXT_REGEX + b"\)Tj",
                        b"()Tj",
                        block,
                    )
                    # remove the url (type 1)
                    block, n_subs2 = re.subn(
                        b"<<\n\/URI \("
                        + DEARXIV_URI_REGEX
                        + b"\)\n\/S /URI\n>>\n",
                        b"",
                        block,
                    )
                    # remove the url (type 2, i.e. Jackson arXiv 0309285v2)
                    block, n_subs3 = re.subn(
                        b"<<\n\/S \/URI\n"
                        + b"/URI \("
                        + DEARXIV_URI_REGEX
                        + b"\)\n>>\n",
                        b"",
                        block,
                    )

                    if n_subs1 or n_subs2:
                        # fix the length of the object stream
                        block = fix_stream_length(block)
                        replaced_arXiv = True
                    new_data.append(block)
                    char_count += len(block)
                    current_obj = []
                elif line in [b"xref\n", b"endobj xref\n"]:
                    if b"endobj" in line and current_obj:
                        current_obj.append(b"endobj\n")
                        block = b"".join(current_obj)
                        new_data.append(block)
                        char_count += len(block)
                        current_obj = []
                        line = b"xref\n"
                    # We found the xref table, record its position and write it
                    # out using our updated indices.
                    startxref = sum(map(len, new_data))
                    new_data.append(line)
                    new_data.append(b"0 %i\n" % (len(xref) + 1))
                    new_data.append(b"0000000000 65535 f \n")
                    for objid in sorted(xref):
                        new_data.append(b"%010d 00000 n \n" % xref[objid])

                    # skip the appropriate number of lines
                    skip_n = len(xref) + 2
                elif current_obj:
                    # If we're recording an object, simply add the line to it
                    current_obj.append(line)
                elif line == b"startxref\n":
                    # Write out our recorded startxref position, skip the old
                    # position.
                    new_data.append(b"startxref\n%i\n" % startxref)
                    skip_n = 1
                else:
                    # Anything else passes through
                    new_data.append(line)
                    char_count += len(line)

        removed_file = basename + "_removed.pdf"
        with open(removed_file, "wb") as fp:
            fp.write(b"".join(new_data))

        output_file = basename + "_dearxiv.pdf"
        self.compress_pdf(removed_file, output_file)

        logger.append("success" if replaced_arXiv else "none found", "info")

        return output_file


def fix_stream_length(block):
    # This fixes the stream length of a block, which is needed after we have
    # removed the arXiv stamp.
    count = 0
    block = block.split(b"\n")
    do_count = False

    for line in block:
        if line.strip(b" ") in [b"stream", b"endstream"]:
            do_count = not do_count
            continue

        if do_count:
            # +1 for the newline character
            count += len(line) + 1

    new_block = []
    for line in block:
        if b" /Length " in line:
            new_block.append(b"<< /Length %i >>" % count)
        else:
            new_block.append(line)

    return b"\n".join(new_block)



# -*- coding: utf-8 -*-

"""Base for the Provider class

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import abc
import os
import shutil
import subprocess
import tempfile
import time

from ..exceptions import _CalledProcessError
from ..log import Logger
from ..pdf_ops import prepare_pdf, blank_pdf, shrink_pdf
from ..utils import (
    assert_file_is_pdf,
    check_pdftool,
    download_url,
    follow_redirects,
    upload_to_remarkable,
)
from ._info import Informer

logger = Logger()


class Provider(metaclass=abc.ABCMeta):
    """ ABC for providers of pdf sources """

    def __init__(
        self,
        verbose=False,
        upload=True,
        debug=False,
        experimental=False,
        crop="left",
        blank=False,
        remarkable_dir="/",
        rmapi_path="rmapi",
        pdftoppm_path="pdftoppm",
        pdftk_path="pdftk",
        qpdf_path="qpdf",
        gs_path="gs",
        css=None,
        font_urls=None,
        cookiejar=None,
    ):
        self.upload = upload
        self.debug = debug
        self.experimental = experimental
        self.remarkable_dir = remarkable_dir
        self.rmapi_path = rmapi_path
        self.pdftoppm_path = pdftoppm_path
        self.pdftk_path = pdftk_path
        self.qpdf_path = qpdf_path
        self.gs_path = gs_path
        self.css = css
        self.font_urls = font_urls
        self.cookiejar = cookiejar

        self.informer = Informer()

        self.pdftool = check_pdftool(self.pdftk_path, self.qpdf_path)

        # wait time to not hit the server too frequently
        self.server_delay = 0

        # disable logging if requested
        if not verbose:
            logger.disable()

        # Define the operations to run on the pdf. Providers can add others.
        self.operations = [("rewrite", self.rewrite_pdf)]
        if crop == "center":
            self.operations.append(("center", self.center_pdf))
        elif crop == "right":
            self.operations.append(("right", self.right_pdf))
        elif crop == "left":
            self.operations.append(("crop", self.crop_pdf))

        if blank:
            self.operations.append(("blank", blank_pdf))
        self.operations.append(("shrink", self.shrink_pdf))

        logger.info("Starting %s provider" % type(self).__name__)

    @staticmethod
    @abc.abstractmethod
    def validate(src):
        """ Validate whether ``src`` is appropriate for this provider """

    @abc.abstractmethod
    def get_abs_pdf_urls(self, src):
        """ Get the url for the HTML page and the PDF file """

    # Wrappers for pdf operations that have additional arguments
    def crop_pdf(self, filepath):
        return prepare_pdf(filepath, "crop", pdftoppm_path=self.pdftoppm_path)

    def center_pdf(self, filepath):
        return prepare_pdf(
            filepath, "center", pdftoppm_path=self.pdftoppm_path
        )

    def right_pdf(self, filepath):
        return prepare_pdf(filepath, "right", pdftoppm_path=self.pdftoppm_path)

    def shrink_pdf(self, filepath):
        return shrink_pdf(filepath, gs_path=self.gs_path)

    def retrieve_pdf(self, pdf_url, filename):
        """ Download pdf from src and save to filename """
        # This must exist so that the LocalFile provider can overwrite it
        download_url(pdf_url, filename, cookiejar=self.cookiejar)

    def compress_pdf(self, in_pdf, out_pdf):
        """ Compress a pdf file, returns subprocess status """
        if self.pdftool == "pdftk":
            status = subprocess.call(
                [self.pdftk_path, in_pdf, "output", out_pdf, "compress"]
            )
        elif self.pdftool == "qpdf":
            status = subprocess.call(
                [
                    self.qpdf_path,
                    "--stream-data=compress",
                    in_pdf,
                    out_pdf,
                ],
                stderr=subprocess.DEVNULL,
            )
        if not status == 0:
            raise _CalledProcessError(
                "%s failed to compress the PDF file." % self.pdftool
            )

    def rewrite_pdf(self, in_pdf, out_pdf=None):
        """Re-write the pdf using Ghostscript

        This helps avoid issues in dearxiv due to nested pdfs.
        """
        if out_pdf is None:
            out_pdf = os.path.splitext(in_pdf)[0] + "-rewrite.pdf"

        status = subprocess.call(
            [
                self.gs_path,
                "-sDEVICE=pdfwrite",
                "-dQUIET",
                "-o",
                out_pdf,
                in_pdf,
            ]
        )
        if not status == 0:
            raise _CalledProcessError(
                "Failed to rewrite the pdf with GhostScript"
            )
        return out_pdf

    def uncompress_pdf(self, in_pdf, out_pdf):
        """ Uncompress a pdf file """

        if self.pdftool == "pdftk":
            status = subprocess.call(
                [
                    self.pdftk_path,
                    in_pdf,
                    "output",
                    out_pdf,
                    "uncompress",
                ]
            )
        elif self.pdftool == "qpdf":
            status = subprocess.call(
                [
                    self.qpdf_path,
                    "--stream-data=uncompress",
                    in_pdf,
                    out_pdf,
                ]
            )
        if not status == 0:
            raise _CalledProcessError(
                "%s failed to uncompress the PDF file." % self.pdftool
            )

    def run(self, src, filename=None):
        # follow_redirects here is needed with library use
        if os.path.exists(src):
            src = src
        elif self.cookiejar is None:
            # NOTE: We assume that if the cookiejar is not None, we are
            # properly redirected.
            src, self.cookiejar = follow_redirects(src)
            time.sleep(self.server_delay)

        # extract page and pdf file urls
        abs_url, pdf_url = self.get_abs_pdf_urls(src)

        # generate nice filename if needed
        clean_filename = filename or self.informer.get_filename(abs_url)
        tmp_filename = "paper.pdf"

        self.initial_dir = os.getcwd()
        with tempfile.TemporaryDirectory(prefix="p2r_") as working_dir:
            os.chdir(working_dir)
            self.retrieve_pdf(pdf_url, tmp_filename)

            assert_file_is_pdf(tmp_filename)

            intermediate_fname = tmp_filename
            for opname, op in self.operations:
                intermediate_fname = op(intermediate_fname)
            shutil.copy(intermediate_fname, clean_filename)

            if self.debug:
                print("Paused in debug mode in dir: %s" % working_dir)
                print("Press enter to exit.")
                return input()

            if self.upload:
                return upload_to_remarkable(
                    clean_filename,
                    remarkable_dir=self.remarkable_dir,
                    rmapi_path=self.rmapi_path,
                )

            target_path = os.path.join(self.initial_dir, clean_filename)
            while os.path.exists(target_path):
                base = os.path.splitext(target_path)[0]
                target_path = base + "_.pdf"
            shutil.move(clean_filename, target_path)
        os.chdir(self.initial_dir)
        return target_path



# -*- coding: utf-8 -*-

from .acm import ACM
from .arxiv import Arxiv
from .citeseerx import CiteSeerX
from .cvf import CVF
from .html import HTML
from .jmlr import JMLR
from .local import LocalFile
from .nature import Nature
from .nber import NBER
from .neurips import NeurIPS
from .openreview import OpenReview
from .pdf_url import PdfUrl
from .pmlr import PMLR
from .pubmed import PubMed
from .sagepub import SagePub
from .science_direct import ScienceDirect
from .semantic_scholar import SemanticScholar
from .springer import Springer
from .tandfonline import TandFOnline

# NOTE: Order matters here, PdfUrl and HTML should be last
providers = [
    ACM,
    Arxiv,
    CiteSeerX,
    CVF,
    JMLR,
    Nature,
    NBER,
    NeurIPS,
    OpenReview,
    PMLR,
    PubMed,
    SagePub,
    ScienceDirect,
    Springer,
    SemanticScholar,
    TandFOnline,
    LocalFile,
    PdfUrl,
    HTML,
]



# -*- coding: utf-8 -*-

"""Operations on PDF files

Author: G.J.J. van den Burg
License: See LICENSE file.
Copyright: 2019, G.J.J. van den Burg

"""


import PyPDF2
import os
import subprocess

from .crop import Cropper
from .log import Logger

logger = Logger()


def prepare_pdf(filepath, operation, pdftoppm_path="pdftoppm"):
    """Prepare pdf by cropping, centering, or right-aligning the flie"""
    logger.info("Preparing PDF using %s operation" % operation)
    prepared_file = os.path.splitext(filepath)[0] + "-prep.pdf"
    cropper = Cropper(filepath, prepared_file, pdftoppm_path=pdftoppm_path)
    if operation == "crop":
        status = cropper.crop(margins=15)
    elif operation == "center":
        status = cropper.center()
    elif operation == "right":
        status = cropper.right()
    else:
        logger.warning("Unknown operation: %s" % operation)
        return filepath
    if not status == 0 or not os.path.exists(prepared_file):
        logger.warning("PDF prepare operation failed")
        return filepath
    return prepared_file


def blank_pdf(filepath):
    """Add blank pages to PDF"""
    logger.info("Adding blank pages")
    input_pdf = PyPDF2.PdfFileReader(filepath)
    output_pdf = PyPDF2.PdfFileWriter()
    for page in input_pdf.pages:
        output_pdf.addPage(page)
        output_pdf.addBlankPage()

    output_file = os.path.splitext(filepath)[0] + "-blank.pdf"
    with open(output_file, "wb") as fp:
        output_pdf.write(fp)
    return output_file


def shrink_pdf(filepath, gs_path="gs"):
    """Shrink the PDF file size using Ghostscript"""
    logger.info("Shrinking pdf file ...")
    size_before = os.path.getsize(filepath)
    output_file = os.path.splitext(filepath)[0] + "-shrink.pdf"
    status = subprocess.call(
        [
            gs_path,
            "-sDEVICE=pdfwrite",
            "-dCompatibilityLevel=1.4",
            "-dPDFSETTINGS=/printer",
            "-dNOPAUSE",
            "-dBATCH",
            "-dQUIET",
            "-sOutputFile=%s" % output_file,
            filepath,
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if not status == 0:
        logger.warning("Failed to shrink the pdf file")
        return filepath
    size_after = os.path.getsize(output_file)
    if size_after > size_before:
        logger.info("Shrinking has no effect for this file, using original.")
        return filepath
    return output_file



# -*- coding: utf-8 -*-

"""Provider for PubMed

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class PubMedInformer(Informer):

    meta_author_key = "citation_authors"

    def _format_authors(self, soup_authors):
        op = lambda x: x[0].split(",")
        return super()._format_authors(soup_authors, sep=" ", idx=-1, op=op)

    def _format_year(self, soup_date):
        if re.match("\w+\ \d{4}", soup_date):
            return soup_date.split(" ")[-1]
        return soup_date.replace(" ", "_")


class PubMed(Provider):

    re_abs = "https?://www.ncbi.nlm.nih.gov/pmc/articles/PMC\d+/?"
    re_pdf = (
        "https?://www.ncbi.nlm.nih.gov/pmc/articles/PMC\d+/pdf/nihms\d+\.pdf"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = PubMedInformer()

    def get_abs_pdf_urls(self, url):
        """Get the pdf and html url from a given PMC url """
        if re.match(self.re_pdf, url):
            idx = url.index("pdf")
            abs_url = url[: idx - 1]
            pdf_url = url
        elif re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url.rstrip("/") + "/pdf"  # it redirects, usually
        else:
            raise URLResolutionError("PMC", url)
        return abs_url, pdf_url

    def validate(src):
        return re.match(PubMed.re_abs, src) or re.match(PubMed.re_pdf, src)



# -*- coding: utf-8 -*-

"""Provider for CiteSeerX

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re
import time

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()


class CiteSeerXInformer(Informer):

    meta_author_key = "citation_authors"
    meta_date_key = "citation_year"

    def _format_authors(self, soup_authors):
        op = lambda x: x[0].split(",")
        return super()._format_authors(soup_authors, sep=" ", idx=-1, op=op)


class CiteSeerX(Provider):

    re_abs = "^https?:\/\/citeseerx.ist.psu.edu\/viewdoc\/summary\?doi=(?P<doi>[0-9\.]+)"
    re_pdf = "^https?:\/\/citeseerx.ist.psu.edu\/viewdoc\/download(\;jsessionid=[A-Z0-9]+)?\?doi=(?P<doi>[0-9\.]+)&rep=rep1&type=pdf"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = CiteSeerXInformer()
        self.server_delay = 30

        # NOTE: This is here because of this:
        # https://github.com/SeerLabs/CiteSeerX/blob/8a62545ffc904f2b41b4ecd30ce91900dc7790f4/src/java/edu/psu/citeseerx/webutils/SimpleDownloadLimitFilter.java#L136
        # The server does not allow hits to the same URL twice within a 30
        # second window. We need to hit the URL more than once to ensure it
        # redirects properly. Waiting is therefore needed.
        logger.info(
            "Waiting 30 seconds so we don't overload the CiteSeerX server."
        )
        time.sleep(30)

    def _get_doi(self, url):
        m = re.match(self.re_abs, url) or re.match(self.re_pdf, url)
        if m:
            return m["doi"]
        raise URLResolutionError(
            "CiteSeerX", url, reason="Failed to retrieve DOI."
        )

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract url from a OpenReview url """
        if re.match(self.re_abs, url):
            abs_url = url
            doi = self._get_doi(abs_url)
            pdf_url = "http://citeseerx.ist.psu.edu/viewdoc/download?doi={doi}&rep=rep1&type=pdf".format(
                doi=doi
            )
        elif re.match(self.re_pdf, url):
            pdf_url = url
            doi = self._get_doi(pdf_url)
            abs_url = "http://citeseerx.ist.psu.edu/viewdoc/summary?doi={doi}".format(
                doi=doi
            )
        else:
            raise URLResolutionError("CiteSeerX", url)
        return abs_url, pdf_url

    def validate(src):
        return re.match(CiteSeerX.re_abs, src) or re.match(
            CiteSeerX.re_pdf, src
        )



# -*- coding: utf-8 -*-

"""Functionality for retrieving paper info
"""

import titlecase
import unidecode
import bs4

from ..utils import clean_string, get_page_with_retry
from ..log import Logger

logger = Logger()


class Informer:
    """Base class for the informers.

    The "informer" class is used to retrieve the title, authors, and year of
    publication of the provided paper.

    This base class provides the main functionality, but because various
    outlets use different conventions to embed author, title, and publication
    year information, we expect that individual providers will subclass this
    class and overwrite some of the methods.
    """

    meta_author_key = "citation_author"
    meta_title_key = "citation_title"
    meta_date_key = "citation_date"

    def __init__(self, title=None, authors=None, year=None):
        self.title = title
        self.authors = authors or []
        self.year = year

    def get_filename(self, abs_url):
        """Generate nice filename using the paper information

        The provided url must be to a HTMl page where this information can be
        found, not to the PDF file itself.
        """
        logger.info("Generating output filename")

        # Retrieve the paper information
        self.get_info(abs_url)

        # we assume that the list of authors is surname only.
        if len(self.authors) > 3:
            authors = self.authors[0] + "_et_al"
        else:
            authors = "_".join(self.authors)
        authors = authors.replace(" ", "_")
        authors = clean_string(authors)

        # Clean the title and make it titlecase
        title = clean_string(self.title)
        title = titlecase.titlecase(title)
        title = title.replace(" ", "_")
        title = clean_string(title)

        year = str(self.year)

        name = authors + "_-_" + title + "_" + year + ".pdf"
        name = unidecode.unidecode(name)
        logger.info("Created filename: %s" % name)
        return name

    def get_info(self, url):
        logger.info("Getting paper info")
        page = get_page_with_retry(url)
        soup = bs4.BeautifulSoup(page, "html.parser")
        self.authors = self.authors or self.get_authors(soup)
        self.title = self.title or self.get_title(soup)
        self.year = self.year or self.get_year(soup)

    ## Title

    def get_title(self, soup):
        meta = soup.find_all("meta", {"name": self.meta_title_key})
        if not meta:
            logger.warning(
                "Couldn't determine title information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        return meta[0]["content"]

    ## Authors

    def _format_authors(self, soup_authors, sep=",", idx=0, op=None):
        op = (lambda x: x) if op is None else op
        # format the author list retrieved by bs4
        return [x.strip().split(sep)[idx].strip() for x in op(soup_authors)]

    def get_authors(self, soup):
        meta = soup.find_all("meta", {"name": self.meta_author_key})
        if not meta:
            logger.warning(
                "Couldn't determine author information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        authors = [x["content"] for x in meta]
        return self._format_authors(authors)

    ## Year

    def _format_year(self, soup_date):
        return soup_date.split("/")[0]

    def get_year(self, soup):
        """ Retrieve the contents of the meta_date_key field and format it """
        meta = soup.find_all("meta", {"name": self.meta_date_key})
        if not meta:
            return ""
        date = meta[0]["content"]
        return self._format_year(date)



# -*- coding: utf-8 -*-

"""
Caller for the command line application
"""

import sys


def main():
    from .ui import main as realmain

    sys.exit(realmain())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Provider for ScienceDirect

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re
import bs4
import urllib
import json

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..log import Logger
from ..utils import get_page_with_retry, follow_redirects

logger = Logger()


class ScienceDirectInformer(Informer):

    meta_date_key = "citation_publication_date"

    def get_authors(self, soup):
        surname_tags = soup.find_all("span", attrs={"class": "text surname"})
        if not surname_tags:
            logger.warning(
                "Couldn't determine author information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        authors = [x.text for x in surname_tags]
        return authors


class ScienceDirect(Provider):

    re_abs = (
        "https?:\/\/www.sciencedirect.com/science/article/pii/[A-Za-z0-9]+"
    )
    re_pdf = "https://pdf.sciencedirectassets.com/\d+/([0-9a-zA-Z\-\.]+)/(?P<data>[0-9a-zA-Z\-\.]+)/main.pdf\?.*"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = ScienceDirectInformer()

    def get_abs_pdf_urls(self, url):
        m1 = re.match(self.re_abs, url)
        m2 = re.match(self.re_pdf, url)
        if m1:
            abs_url = url
            pdf_url = self._get_pdf_url(abs_url)
        elif m2:
            pdf_url = url
            data = m2.group("data")
            paper_id = data.split("-")[-1]
            abs_url = (
                f"https://www.sciencedirect.com/science/article/pii/{paper_id}"
            )
        else:
            raise URLResolutionError("ScienceDirect", url)
        return abs_url, pdf_url

    def _get_pdf_url(self, url):
        page = get_page_with_retry(url)
        soup = bs4.BeautifulSoup(page, "html.parser")

        # For open access (and maybe behind institution?) the full text pdf url
        # is currently in the json payload of a script tag.
        scripts = soup.find_all("script", attrs={"data-iso-key": "_0"})
        if not scripts:
            raise URLResolutionError("ScienceDirect", url)
        json_data = scripts[0].string
        data = json.loads(json_data)
        if not "article" in data:
            raise URLResolutionError("ScienceDirect", url)
        data = data["article"]
        if not "pdfDownload" in data:
            raise URLResolutionError("ScienceDirect", url)
        data = data["pdfDownload"]
        if not "linkToPdf" in data:
            raise URLResolutionError("ScienceDirect", url)
        link = data["linkToPdf"]
        tmp_url = urllib.parse.urljoin("https://sciencedirect.com/", link)

        # tmp_url gives a page with a ten second wait or a direct url, we need
        # the direct url
        page = get_page_with_retry(tmp_url)
        soup = bs4.BeautifulSoup(page, "html.parser")
        noscript = soup.find_all("noscript")
        if not noscript:
            raise URLResolutionError("ScienceDirect", url)
        a = noscript[0].find_all("a")
        if not a:
            raise URLResolutionError("ScienceDirect", url)
        pdf_url = a[0].get("href")
        return pdf_url

    def validate(src):
        return re.match(ScienceDirect.re_abs, src) or re.match(
            ScienceDirect.re_pdf, src
        )



# -*- coding: utf-8 -*-

"""Provider for Taylor and Francis Online

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()


class TandFOnlineInformer(Informer):
    meta_title_key = "dc.Title"
    meta_author_key = "dc.Creator"
    meta_date_key = "dc.Date"

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)

    def _format_year(self, soup_date):
        return soup_date.strip().split(" ")[-1].strip()


class TandFOnline(Provider):

    re_abs = "^https?://\w+.tandfonline.com/doi/(full|abs)/(?P<doi>\d+\.\d+/\d+\.\d+\.\d+)"
    re_pdf = "^https?://\w+.tandfonline.com/doi/(full|pdf)/(?P<doi>\d+\.\d+/\d+\.\d+\.\d+)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = TandFOnlineInformer()

    def _get_doi(self, url):
        m = re.match(self.re_abs, url) or re.match(self.re_pdf, url)
        if m:
            return m["doi"]
        raise URLResolutionError(
            "TandFOnline", url, reason="Failed to retrieve DOI."
        )

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            doi = self._get_doi(url)
            pdf_url = "https://www.tandfonline.com/doi/pdf/{doi}?needAccess=true".format(
                doi=doi
            )
        elif re.match(self.re_pdf, url):
            doi = self._get_doi(url)
            pdf_url = "https://www.tandfonline.com/doi/pdf/{doi}?needAccess=true".format(
                doi=doi
            )
            # full redirects to abs if we don't have access
            abs_url = "https://www.tandfonline.com/doi/full/{doi}".format(
                doi=doi
            )
        else:
            raise URLResolutionError("TandFOnline", url)
        return abs_url, pdf_url

    def validate(src):
        m = re.match(TandFOnline.re_abs, src) or re.match(
            TandFOnline.re_pdf, src
        )
        return not m is None



# -*- coding: utf-8 -*-

"""Exceptions for paper2remarkable

"""

from . import GITHUB_URL

GH_MSG = (
    "\n\nIf you think this might be a bug, please raise an issue on "
    "GitHub at:\n{url}\n".format(url=GITHUB_URL)
)


class Error(Exception):
    """Base class for exceptions in p2r."""


class URLResolutionError(Error):
    """Exception raised when the abstract/pdf urls can't be resolved.

    Attributes
    ----------
    provider : str
        Name of the provider where the error occurred

    url : str
        Original url provided

    """

    def __init__(self, provider, url, reason=None):
        self.provider = provider
        self.url = url
        self.reason = None

    def __str__(self):
        msg = "ERROR: Couldn't figure out {provider} URLs from provided url: {url}".format(
            provider=self.provider, url=self.url
        )
        if self.reason:
            msg += "\nReason: {reason}".format(reason=self.reason)
        msg += GH_MSG
        return msg


class FilenameMissingError(Error):
    """Exception raised for providers that need a filename to be provided"""

    def __init__(self, provider, url, reason=None):
        self.provider = provider
        self.url = url
        self.reason = reason

    def __str__(self):
        msg = "ERROR: Couldn't determine a filename from {url} for provider {provider}".format(
            provider=self.provider, url=self.url
        )
        if self.reason:
            msg += "\nReason: {reason}".format(reason=self.reason)
        msg += GH_MSG
        return msg


class FileTypeError(Error):
    """Exception raised when we have a mismatch in filetype."""

    def __init__(self, filename, filetype):
        self.filename = filename
        self.filetype = filetype

    def __str__(self):
        msg = "ERROR: File {filename} isn't of type {filetype}.".format(
            filename=self.filename, filetype=self.filetype
        )
        msg += GH_MSG
        return msg


class RemarkableError(Error):
    """Exceptions raised when interacting with the reMarkable fails."""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        msg = "ERROR: {message}".format(message=self.message)
        msg += GH_MSG
        return msg


class _CalledProcessError(Error):
    """Exception raised when subprocesses fail.  """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        msg = "ERROR: {message}".format(message=self.message)
        msg += GH_MSG
        return msg


class NoPDFToolError(Error):
    """Exception raised when neither pdftk or qpdf is found."""

    def __init__(self):
        pass

    def __str__(self):
        msg = (
            "ERROR: Neither pdftk or qpdf could be found. Install "
            "either of these or ensure that they can be found using "
            "the --pdftk or --qpdf options."
        )
        msg += GH_MSG
        return msg


class UnidentifiedSourceError(Error):
    """Exception raised when the input is neither a local file nor a url """

    def __str__(self):
        msg = (
            "ERROR: Couldn't figure out what source you mean. If it's a "
            "local file, please make sure it exists."
        )
        msg += GH_MSG
        return msg


class InvalidURLError(Error):
    """Exception raised when no provider can handle a url source """

    def __str__(self):
        msg = (
            "ERROR: Input URL is not valid, no provider can handle "
            "this source."
        )
        msg += GH_MSG
        return msg



# -*- coding: utf-8 -*-

"""Provider for HTML documents

This provider is a little bit special, in that it isn't simply pulling an 
academic paper from a site, but instead aims to pull a HTML article.

Author: G.J.J. van den Burg
License: See LICENSE file.
Copyright: 2020, G.J.J. van den Burg

"""

import html2text
import markdown
import re
import readability
import titlecase
import unidecode
import urllib
import weasyprint

from ._base import Provider
from ._info import Informer

from ..utils import (
    clean_string,
    get_page_with_retry,
    get_content_type_with_retry,
)
from ..log import Logger

logger = Logger()

CSS = """
@page { size: 702px 936px; margin: 1in; }
a { color: black; }
img { display: block; margin: 0 auto; text-align: center; max-width: 70%; max-height: 300px; }
p, li { font-size: 10pt; font-family: 'EB Garamond'; hyphens: auto; text-align: justify; }
h1,h2,h3 { font-family: 'Noto Serif'; }
h1 { font-size: 26px; }
h2 { font-size: 18px; }
h3 { font-size: 14px; }
blockquote { font-style: italic; }
pre { font-family: 'Inconsolata'; padding-left: 2.5%; background: #efefef; }
code { font-family: 'Inconsolata'; font-size: .7rem; background: #efefef; }
"""

# NOTE: For some reason, Weasyprint no longer accepts the @import statement in
# the CSS to load the fonts. This may have to do with recent changes they've
# introduced. Providing the font urls separately does seem to work.
FONT_URLS = [
    "https://fonts.googleapis.com/css2?family=EB+Garamond&family=Noto+Serif&family=Inconsolata"
]


def url_fetcher(url):
    if url.startswith("//"):
        url = "https:" + url
    elif url.startswith("file:///"):
        url = "https:" + url[len("file:/") :]
    return weasyprint.default_url_fetcher(url)


def make_readable(request_html):
    """Use an extraction method to get the main article html

    This function checks if ReadabiliPy is installed with NodeJS support, as
    that generally yields better results. If that is not available, it falls
    back on readability.
    """

    have_readabilipy_js = False
    try:
        import readabilipy

        have_readabilipy_js = readabilipy.simple_json.have_node()
    except ImportError:
        pass

    if have_readabilipy_js:
        logger.info("Converting HTML using Readability.js")
        article = readabilipy.simple_json_from_html_string(
            request_html, use_readability=True
        )
        title = article["title"]
        raw_html = article["content"]
    else:
        logger.info("Converting HTML using readability")
        doc = readability.Document(request_html)
        title = doc.title()
        raw_html = doc.summary(html_partial=True)
    return title, raw_html


class ImgProcessor(markdown.treeprocessors.Treeprocessor):
    def __init__(self, base_url, *args, **kwargs):
        self._base_url = base_url
        super().__init__(*args, **kwargs)

    def run(self, root):
        """ Ensure all img src urls are absolute """
        for img in root.iter("img"):
            img.attrib["src"] = urllib.parse.urljoin(
                self._base_url, img.attrib["src"]
            )
            img.attrib["src"] = img.attrib["src"].rstrip("/")


class HTMLInformer(Informer):
    def __init__(self):
        super().__init__()
        self._cached_title = None
        self._cached_article = None

    def get_filename(self, abs_url):
        request_html = get_page_with_retry(abs_url, return_text=True)
        title, article = make_readable(request_html)

        self._cached_title = title
        self._cached_article = article

        # Clean the title and make it titlecase
        title = clean_string(title)
        title = titlecase.titlecase(title)
        title = title.replace(" ", "_")
        title = clean_string(title)
        name = title.strip("_") + ".pdf"
        name = unidecode.unidecode(name)
        logger.info("Created filename: %s" % name)
        return name


class HTML(Provider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = HTMLInformer()

    def get_abs_pdf_urls(self, url):
        return url, url

    def fix_lazy_loading(self, article):
        if not self.experimental:
            return article

        # This attempts to fix sites where the image src element points to a
        # placeholder and the data-src attribute contains the url to the actual
        # image. Note that results may differ between readability and
        # Readability.JS
        regex = '<img src="(?P<src>.*?)" (?P<rest1>.*) data-src="(?P<datasrc>.*?)" (?P<rest2>.*?)>'
        sub = '<img src="\g<datasrc>" \g<rest1> \g<rest2>>'

        article, nsub = re.subn(regex, sub, article, flags=re.MULTILINE)
        if nsub:
            logger.info(
                f"[experimental] Attempted to fix lazy image loading ({nsub} times). "
                "Please report bad results."
            )
        return article

    def preprocess_html(self, pdf_url, title, article):
        article = self.fix_lazy_loading(article)

        h2t = html2text.HTML2Text()
        h2t.wrap_links = False
        text = h2t.handle(article)

        # Add the title back to the document
        article = "# {title}\n\n{text}".format(title=title, text=text)

        # Convert to html, fixing relative image urls.
        md = markdown.Markdown()
        md.treeprocessors.register(ImgProcessor(pdf_url), "img", 10)
        html_article = md.convert(article)
        return html_article

    def retrieve_pdf(self, pdf_url, filename):
        """Turn the HTML article in a clean pdf file

        This function takes the following steps:

        1. Pull the HTML page using requests, if not done in Informer
        2. Extract the article part of the page using readability/readabiliPy
        3. Convert the article HTML to markdown using html2text
        4. Convert the markdown back to HTML (done to sanitize the HTML)
        4. Convert the HTML to PDF, pulling in images where needed
        5. Save the PDF to the specified filename.
        """
        if self.informer._cached_title and self.informer._cached_article:
            title = self.informer._cached_title
            article = self.informer._cached_article
        else:
            request_html = get_page_with_retry(pdf_url, return_text=True)
            title, article = make_readable(request_html)

        html_article = self.preprocess_html(pdf_url, title, article)

        if self.debug:
            with open("./paper.html", "w") as fp:
                fp.write(html_article)

        html = weasyprint.HTML(string=html_article, url_fetcher=url_fetcher)
        css = CSS if self.css is None else self.css
        font_urls = FONT_URLS if self.font_urls is None else self.font_urls
        style = weasyprint.CSS(string=css)
        html.write_pdf(filename, stylesheets=[style] + font_urls)

    def validate(src):
        # first check if it is a valid url
        parsed = urllib.parse.urlparse(src)
        if not all([parsed.scheme, parsed.netloc, parsed.path]):
            return False
        # next, get the header and check the content type
        ct = get_content_type_with_retry(src)
        if ct is None:
            return False
        return ct.startswith("text/html")



# -*- coding: utf-8 -*-

"""Provider for ACM

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError
from ..log import Logger

logger = Logger()


class ACMInformer(Informer):
    meta_author_key = "citation_authors"

    def get_title(self, soup):
        target = soup.find("h1", {"class": "citation__title"})
        return target.text

    def get_authors(self, soup):
        authors = [
            span.find("a").text
            for span in soup.find_all("span", {"class": "auth-name"})
        ]
        return self._format_authors(authors)

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)

    def get_year(self, soup):
        date = soup.find("span", {"class": "epub-section__date"})
        return self._format_year(date.text)

    def _format_year(self, soup_date):
        return soup_date.strip().split(" ")[-1].strip()


class ACM(Provider):

    re_abs = "^https?://dl.acm.org/doi/(?P<doi>\d+\.\d+/\d+\.\d+)"
    re_pdf = "^https?://dl.acm.org/doi/pdf/(?P<doi>\d+\.\d+/\d+\.\d+)\?download=true"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = ACMInformer()

    def _get_doi(self, url):
        m = re.match(self.re_abs, url) or re.match(self.re_pdf, url)
        if m:
            return m["doi"]
        raise URLResolutionError("ACM", url, reason="Failed to retrieve DOI.")

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            doi = self._get_doi(url)
            pdf_url = "https://dl.acm.org/doi/pdf/{doi}?download=true".format(
                doi=doi
            )
        elif re.match(self.re_pdf, url):
            pdf_url = url
            doi = self._get_doi(url)
            abs_url = "https://dl.acm.org/doi/{doi}".format(doi=doi)
        else:
            raise URLResolutionError("ACM", url)
        return abs_url, pdf_url

    def validate(src):
        m = re.match(ACM.re_abs, src) or re.match(ACM.re_pdf, src)
        return not m is None



# -*- coding: utf-8 -*-

"""Command line interface

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import argparse
import copy
import os
import sys
import yaml

from . import __version__, GITHUB_URL

from .exceptions import UnidentifiedSourceError, InvalidURLError
from .providers import providers, LocalFile
from .utils import follow_redirects, is_url


def build_argument_parser():
    parser = argparse.ArgumentParser(
        description="Paper2reMarkable version %s" % __version__
    )
    parser.add_argument(
        "-b",
        "--blank",
        help="Add a blank page after every page of the PDF",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--center",
        help="Center the PDF on the page, instead of left align",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="debug mode, doesn't upload to reMarkable",
        action="store_true",
    )
    parser.add_argument(
        "-e",
        "--experimental",
        help="enable experimental features",
        action="store_true",
    )
    parser.add_argument(
        "-n",
        "--no-upload",
        help="don't upload to reMarkable, save the output in current directory",
        action="store_true",
    )
    parser.add_argument(
        "-p",
        "--remarkable-path",
        help="directory on reMarkable to put the file (created if missing, default: /)",
        dest="remarkable_dir",
        default="/",
    )
    parser.add_argument(
        "-r",
        "--right",
        help="Right align so the menu doesn't cover it",
        action="store_true",
    )
    parser.add_argument(
        "-k", "--no-crop", help="Don't crop the pdf file", action="store_true"
    )
    parser.add_argument(
        "-v", "--verbose", help="be verbose", action="store_true"
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Show version and exit",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "-f",
        "--filename",
        help="Filename to use for the file on reMarkable",
        action="append",
    )
    parser.add_argument(
        "--gs", help="path to gs executable (default: gs)", default=None
    )
    parser.add_argument(
        "--pdftoppm",
        help="path to pdftoppm executable (default: pdftoppm)",
        default=None,
    )
    parser.add_argument(
        "--pdftk",
        help="path to pdftk executable (default: pdftk)",
        default=None,
    )
    parser.add_argument(
        "--qpdf",
        help="path to qpdf executable (default: qpdf)",
        default=None,
    )
    parser.add_argument(
        "--rmapi",
        help="path to rmapi executable (default: rmapi)",
        default=None,
    )
    parser.add_argument(
        "--css", help="path to custom CSS file for HTML output", default=None
    )
    parser.add_argument(
        "--font-urls",
        help="path to custom font urls file for HTML output",
        default=None,
    )
    parser.add_argument(
        "-C",
        "--config",
        help="path to config file (default: ~/.paper2remarkable.yml)",
        default=None,
    )
    parser.add_argument(
        "input",
        help="One or more URLs to a paper or paths to local PDF files",
        nargs="+",
    )
    return parser


def parse_args():
    parser = build_argument_parser()
    return parser.parse_args()


def exception(msg):
    print("ERROR: " + msg, file=sys.stderr)
    print("Error occurred. Exiting.", file=sys.stderr)
    print("", file=sys.stderr)
    print(
        "If you think this might be a bug, please raise an issue on GitHub: %s"
        % GITHUB_URL,
        file=sys.stderr,
    )
    print("", file=sys.stderr)
    raise SystemExit(1)


def choose_provider(cli_input):
    """Choose the provider to use for the given source

    This function first tries to check if the input is a local file, by
    checking if the path exists. Next, it checks if the input is a "valid" url
    using a regex test. If it is, the registered provider classes are checked
    to see which provider can handle this url.

    Returns
    -------
    provider : class
        The class of the provider than can handle the source. A subclass of the
        Provider abc.

    new_input : str
        The updated input to the provider. This only has an effect for the url
        providers, where this will be the url after following all redirects.

    cookiejar : dict or requests.RequestsCookieJar
        Cookies picked up when following redirects. These are needed for some
        providers to ensure later requests have the right cookie settings.

    Raises
    ------
    UnidentifiedSourceError
        Raised when the input is neither an existing local file nor a valid url

    InvalidURLError
        Raised when the input *is* a valid url, but no provider can handle it.

    """
    provider = cookiejar = None
    if LocalFile.validate(cli_input):
        # input is a local file
        new_input = cli_input
        provider = LocalFile
    elif is_url(cli_input):
        # input is a url
        new_input, cookiejar = follow_redirects(cli_input)
        provider = next((p for p in providers if p.validate(new_input)), None)
    else:
        # not a proper URL or non-existent file
        raise UnidentifiedSourceError

    if provider is None:
        raise InvalidURLError

    return provider, new_input, cookiejar


def load_config(path=None):
    if path is None:
        path = os.path.join(os.path.expanduser("~"), ".paper2remarkable.yml")
    if not os.path.exists(path):
        return None
    with open(path, "r") as fp:
        config = yaml.safe_load(fp)
    return config


def merge_options(args, config=None):
    # command line arguments always overwrite config
    config = {} if config is None else config

    opts = copy.deepcopy(config)
    opts.setdefault("core", {})
    opts.setdefault("system", {})
    opts.setdefault("html", {})

    def set_bool(d, key, value, invert=False):
        if value:
            d[key] = True ^ invert
        elif not key in d:
            d[key] = False ^ invert

    def set_path(d, key, value):
        if not value is None:
            d[key] = value
        elif not key in d:
            d[key] = key

    set_bool(opts["core"], "blank", args.blank)
    set_bool(opts["core"], "verbose", args.verbose)
    set_bool(opts["core"], "upload", args.no_upload, invert=True)
    set_bool(opts["core"], "experimental", args.experimental)

    if args.center:
        opts["core"]["crop"] = "center"
    elif args.right:
        opts["core"]["crop"] = "right"
    elif args.no_crop:
        opts["core"]["crop"] = "none"
    elif not "crop" in opts["core"]:
        opts["core"]["crop"] = "left"

    set_path(opts["system"], "gs", args.gs)
    set_path(opts["system"], "pdftoppm", args.pdftoppm)
    set_path(opts["system"], "pdftk", args.pdftk)
    set_path(opts["system"], "qpdf", args.qpdf)
    set_path(opts["system"], "rmapi", args.rmapi)

    if args.css and os.path.exists(args.css):
        with open(args.css, "r") as fp:
            contents = fp.read()
        opts["html"]["css"] = contents
    elif not "css" in opts["html"]:
        opts["html"]["css"] = None

    if args.font_urls and os.path.exists(args.font_urls):
        with open(args.font_urls, "r") as fp:
            urls = [l.strip() for l in fp.readlines()]
        opts["html"]["font_urls"] = urls
    elif not "font_urls" in opts["html"]:
        opts["html"]["font_urls"] = None

    return opts


def set_excepthook(debug):
    sys_hook = sys.excepthook

    def exception_handler(exception_type, value, traceback):
        if debug:
            sys_hook(exception_type, value, traceback)
        else:
            print(value, file=sys.stderr)

    sys.excepthook = exception_handler


def main():
    args = parse_args()
    set_excepthook(args.debug)

    if args.center and args.right:
        exception("Can't center and right align at the same time!")

    if args.center and args.no_crop:
        exception("Can't center and not crop at the same time!")

    if args.right and args.no_crop:
        exception("Can't right align and not crop at the same time!")

    if args.filename and not len(args.filename) == len(args.input):
        exception(
            "When providing --filename and multiple inputs, their number must match."
        )

    config = load_config(path=args.config)
    options = merge_options(args, config=config)

    filenames = (
        [None] * len(args.input) if not args.filename else args.filename
    )

    for cli_input, filename in zip(args.input, filenames):
        provider, new_input, cookiejar = choose_provider(cli_input)
        prov = provider(
            verbose=options["core"]["verbose"],
            upload=options["core"]["upload"],
            debug=args.debug,
            experimental=options["core"]["experimental"],
            crop=options["core"]["crop"],
            blank=options["core"]["blank"],
            remarkable_dir=args.remarkable_dir,
            rmapi_path=options["system"]["rmapi"],
            pdftoppm_path=options["system"]["pdftoppm"],
            pdftk_path=options["system"]["pdftk"],
            qpdf_path=options["system"]["qpdf"],
            gs_path=options["system"]["gs"],
            css=options["html"]["css"],
            font_urls=options["html"]["font_urls"],
            cookiejar=cookiejar,
        )
        prov.run(new_input, filename=filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os

from setuptools import find_packages, setup

# Package meta-data.
AUTHOR = "Gertjan van den Burg"
DESCRIPTION = "Easily download an academic paper and send it to the reMarkable"
EMAIL = "gertjanvandenburg@gmail.com"
LICENSE = "MIT"
LICENSE_TROVE = "License :: OSI Approved :: MIT License"
NAME = "paper2remarkable"
REQUIRES_PYTHON = ">=3.5.0"
URL = "https://github.com/GjjvdBurg/paper2remarkable"
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "PyPDF2>=1.26",
    "beautifulsoup4>=4.8",
    "html2text>=2020.1.16",
    "markdown>=3.1.1",
    "pdfplumber>=0.5",
    "pyyaml>=5.1",
    "readability-lxml>=0.7.1",
    "regex>=2018.11",
    "requests>=2.21",
    "titlecase>=0.12",
    "unidecode>=1.1",
    "weasyprint>=51",
]

full_require = ["readabilipy"]
docs_require = []
test_require = ["green"]
dev_require = []

# What packages are optional?
EXTRAS = {
    "full": full_require,
    "docs": docs_require,
    "test": test_require + full_require,
    "dev": docs_require + test_require + dev_require + full_require,
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    data_files=[("man/man1", ["p2r.1"])],
    license=LICENSE,
    ext_modules=[],
    entry_points={"console_scripts": ["p2r = paper2remarkable.__main__:main"]},
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        LICENSE_TROVE,
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

"""Provider for local files

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import os
import shutil

from ._base import Provider
from ._info import Informer


class LocalFileInformer(Informer):
    def get_filename(self, abs_url):
        return os.path.basename(abs_url)


class LocalFile(Provider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = LocalFileInformer()

    def get_abs_pdf_urls(self, url):
        # The 'url' is the path to the local file. We use this as abs_url and
        # pdf_url.
        return url, url

    def validate(src):
        return os.path.exists(src)

    def retrieve_pdf(self, pdf_url, filename):
        source = os.path.join(self.initial_dir, pdf_url)
        shutil.copy(source, filename)



# -*- coding: utf-8 -*-

"""Provider for Nature

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class NatureInformer(Informer):

    meta_date_key = "citation_online_date"

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class Nature(Provider):

    re_abs = "^https://www.nature.com/articles/s[a-z0-9\-]+$"
    re_pdf = "^https://www.nature.com/articles/s[a-z0-9\-]+\.pdf$"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = NatureInformer()

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url + ".pdf"
        elif re.match(self.re_pdf, url):
            pdf_url = url
            abs_url = url.replace(".pdf", "")
        else:
            raise URLResolutionError("Nature", url)
        return abs_url, pdf_url

    def validate(src):
        return re.match(Nature.re_abs, src) or re.match(Nature.re_pdf, src)



# -*- coding: utf-8 -*-

"""Just a simple logger

Author: G.J.J. van den Burg
License: See LICENSE file.
Copyright: 2019, G.J.J. van den Burg

"""

# NOTE: I know about the logging module, but this was easier because one of the
# dependencies was using that and it interfered with our logging. The logger
# class below is obviously not thread-safe and is very simple.

import datetime
import sys


class Singleton(type):
    # https://stackoverflow.com/q/6760685
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls]


class Logger(metaclass=Singleton):
    def __init__(self):
        self.enabled = True

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def _log(self, msg, mode, end="\n", add_prefix=True):
        if not self.enabled:
            return
        if not mode in ("info", "warn"):
            raise ValueError("Unknown logging mode: %s" % mode)
        file = sys.stdout if mode == "info" else sys.stderr
        if add_prefix:
            now = datetime.datetime.now()
            nowstr = now.strftime("%Y-%m-%d %H:%M:%S")
            prefix = "%s - %s - " % (nowstr, mode.upper())
        else:
            prefix = ""
        print("%s%s" % (prefix, msg), end=end, file=file)
        file.flush()

    def info(self, msg, end="\n"):
        self._log(msg, "info", end=end)

    def warning(self, msg, end="\n"):
        self._log(msg, "warn", end=end)

    def append(self, msg, mode, end="\n"):
        self._log(msg, mode, end=end, add_prefix=False)



# -*- coding: utf-8 -*-

"""Provider for PMLR

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import re

from ._base import Provider
from ._info import Informer
from ..exceptions import URLResolutionError


class PMLRInformer(Informer):

    meta_date_key = "citation_publication_date"

    def _format_authors(self, soup_authors):
        return super()._format_authors(soup_authors, sep=" ", idx=-1)


class PMLR(Provider):

    re_abs_1 = "https?://proceedings.mlr.press/v\d+/[\w\-\w]+\d+.html"
    re_pdf_1 = "https?://proceedings.mlr.press/v\d+/[\w\-\w]+\d+.pdf"

    re_abs_2 = "https?://proceedings.mlr.press/v\d+/[\w\-\w]+\d+\w?.html"
    re_pdf_2 = "https?://proceedings.mlr.press/v\d+/(?P<ref>[\w\-\w]+\d+\w?)/(?P=ref).pdf"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = PMLRInformer()

    def get_abs_pdf_urls(self, url):
        """ Get the pdf and abstract url from a OpenReview url """
        if re.match(self.re_abs_1, url):
            abs_url = url
            pdf_url = url.replace(".html", ".pdf")
        elif re.match(self.re_pdf_1, url):
            abs_url = url.replace(".pdf", ".html")
            pdf_url = url
        elif re.match(self.re_abs_2, url):
            abs_url = url
            parts = url.split("/")
            authoridx = parts[-1].split(".")[0]
            pdf_url = "/".join(parts[:-1]) + "/%s/%s.pdf" % (
                authoridx,
                authoridx,
            )
        elif re.match(self.re_pdf_2, url):
            parts = url.split("/")
            abs_url = "/".join(parts[:-1]) + ".html"
            pdf_url = url
        else:
            raise URLResolutionError("PMLR", url)
        return abs_url, pdf_url

    def validate(src):
        return (
            re.fullmatch(PMLR.re_abs_1, src)
            or re.fullmatch(PMLR.re_pdf_1, src)
            or re.fullmatch(PMLR.re_abs_2, src)
            or re.fullmatch(PMLR.re_pdf_2, src)
        )



# -*- coding: utf-8 -*-

"""Provider for generic PDF url

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import urllib

from ._base import Provider
from ._info import Informer

from .. import GITHUB_URL
from ..exceptions import FilenameMissingError
from ..log import Logger
from ..utils import get_content_type_with_retry

logger = Logger()


class PdfUrlInformer(Informer):
    def get_filename(self, abs_url):
        # try to get a nice filename by parsing the url
        parsed = urllib.parse.urlparse(abs_url)
        path_parts = parsed.path.split("/")
        if not path_parts:
            raise FilenameMissingError(
                provider="PdfUrl",
                url=abs_url,
                reason="No URL parts",
            )

        filename = path_parts[-1]
        if not filename.endswith(".pdf"):
            raise FilenameMissingError(
                provider="PdfUrl",
                url=abs_url,
                reason="URL path didn't end in .pdf",
            )
        logger.warning(
            "Using filename {filename} extracted from url. "
            "You might want to provide a nicer one using --filename "
            "or request this paper source to be added "
            "(see: {github}).".format(filename=filename, github=GITHUB_URL)
        )
        return filename


class PdfUrl(Provider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = PdfUrlInformer()

    def get_abs_pdf_urls(self, url):
        return (url, url)

    def validate(src):
        # first check if it is a valid url
        parsed = urllib.parse.urlparse(src)
        if not all([parsed.scheme, parsed.netloc, parsed.path]):
            return False
        # next, get the header and check the content type
        ct = get_content_type_with_retry(src)
        if ct is None:
            return False
        return ct.startswith("application/pdf")
